local TaskManager = require "TaskManager"
local missionutil = require "missionutil"
local vec3 = require "vec3"
local transf = require "transf"
local jj = 0
local expcount = 0

local function makeIntroTask(taskParams, doneFn, label)
	
	local info = {
		name = _("Überblick"),
		paragraphs = {
			{ text = _(
            "Es ist August im Jahre 1872 und der Bau des Gotthardtunnels kann beginnen! Keine Ahnung, warum ausgerechnet wir den Zuschlag für dieses Monsterprojekt erhalten haben – ob es daran liegt, dass wir die einzigen Schweizer waren, die sich beworben haben? Oder daran, dass unsere Offerte so lächerlich tief war, dass kein Konkurrent hatte mithalten können?\n\n\z
			Fakt ist: wir haben acht Jahre Zeit und ein ziemlich knappes Budget, um den längsten Tunnel der Welt zu bauen. Und wir haben noch nie zuvor einen Tunnel gebaut... Also nichts wie los!")
			},
			{ type = "HINT", text = _("Falls Du nicht zurecht kommst, empfiehlt es sich, zuerst das Tutorial zu spielen. Dieses kann über das Hauptmenü gestartet werden.")  
			},
		},
		voiceOver = "europe_01_02.wav",
	}
	
	local function fn(taskManager, task)
		task.setCompleted()
		for i=1, #doneFn do
			doneFn[i](taskManager)
		end
	end

	info.options = { { label or _("Ok!"), fn } }
	info.optionsRightAlign = true

	local task = missionutil.makeTask(info)

	task.update = function(taskManager)
		
	end
	
	task.setProgressNone()

	return task
end

local function makeDeliverTools(taskParams, doneFn)

	local function complete(taskManager, task)
		task.setCompleted()
		missionutil.startTask(taskManager,"connectTunnel")
	end
	
	local info = {
		title = _("Die Gotthardbahn"),
		name = _("Werkzeuge liefern"),
		paragraphs = {
			{ text = _("Schon drei Eisenbahnlinien überqueren die Alpen. Doch wenn unser kühnes Projekt dereinst vollendet ist, wird die Gotthardbahn die schnellste und direkteste Verbindung von Nord nach Süd sein.\n\n\z
			Der harte Gneis des Gotthardmassivs türmt sich vor uns empor. Wir haben genügend Arbeiter – zerlumpte aber kräftige Italiener, die sich für ein Butterbrot verdingt haben. Doch mit blossen Händen graben wir uns nicht durch den Berg... Wir müssen die nötigen Gerätschaften zur Baustelle am Nordportal befördern.")
			}, 
			{ type = "TASK", text = _("Liefere die ersten Werkzeuge zur [link=${place}]Tunnelbaustelle[/link].") % { place = taskParams.constructionSiteConstr }
			},
			{ type = "HINT", text = _("Baue eine LKW-Station bei [link=${toolsFactory}][/link].") %
				{ 
					tunnel = taskParams.constructionSiteConstr,
					toolsFactory = taskParams.toolsFactory,
				}
			},
			{ type = "HINT", text = _("Errichte ein Depot und kaufe dort Pferdefuhrwerke, die Werkzeuge laden können.") %
				{ 
					industry0 = taskParams.industryIronOre,
					industry1 = taskParams.industryToolsFactory,
					town0 	  = taskParams.townOgden,
					town1 	  = taskParams.townEvanston
				}
			},
			{ type = "HINT", text = _("Erstelle anschliessend eine Linie zwischen den beiden LKW-Stationen und setze die Pferdefuhrwerke darauf ein.")
			},			
        },
		camera = { game.interface.getEntity(taskParams.constructionSite).position[1], game.interface.getEntity(taskParams.constructionSite).position[2], 1000.0 },
		voiceOver = "europe_01_03.wav",
		options = {
			{ "Debug: Skip", complete },
		},
	}
	local task = missionutil.makeTask(info)
		
	task.init = function(taskManager, data)
	
	end
	
	task.update = function(taskManager)
		if task.isCompleted() then return end
				
		local toolsCosumed = game.interface.getEntity(taskParams.constructionSite).itemsConsumed._sum or 0
				
		if toolsCosumed > 0 then
			complete(taskManager, task)
		end
		
				
	end
	
	return task
end


local function makeConnectTunnel(taskParams, doneFn)

	local function complete(taskManager, task)
		task.setCompleted()
		missionutil.startTask(taskManager,"startConstruction")
		taskManager:setZone("trackEndAltdorf")
		taskManager:setZone("stationTunnel")
	end
	
	local info = {
		title = _("Die Gotthardbahn"),
		name = _("Tunnelbaustelle anschliessen"),
		paragraphs = {
			{ text = _("Jetzt wird der Berg ausgehöhlt. Faszinierend, wie viel Schutt da zusammenkommt. Mit dem Pferdefuhrwerk allein ist der unmöglich wegzuschaffen...\n\n\z
			Wir brauchen eine ordentliche Bahnlinie, damit wir die riesigen Mengen an Aushub von der Baustelle abtransportieren können.")
			}, 
			{ type = "TASK", text = _("Verbinde [link=${stationTunnel}][/link] und [link=${stationDump}][/link] mit Gleisen") % { stationTunnel = taskParams.stationTunnel, stationDump = taskParams.stationDump }
			},
			{ type = "HINT", text = _("Schliesse die Lücke zwischen den markierten Gleisenden.")
			},
			{ type = "HINT", text = _("Blende die Höhenlinien über das Overlaymenü ein und orientiere Dich am Gelände, um ohne Tunnel und Brücken auszukommen. Das spart Kosten!")
			},	
			{ type = "HINT", text = _("Wenn Du die Gleise in kleinen Schritten baust, passen sie sich besser an das Gelände an.")
			},	
        },
		camera = { (game.interface.getEntity(taskParams.stationTunnel).position[1] + game.interface.getEntity(taskParams.stationDump).position[1]) / 2, (game.interface.getEntity(taskParams.stationTunnel).position[2] + game.interface.getEntity(taskParams.stationDump).position[2]) / 2, 2000.0 },
		voiceOver = "europe_01_04.wav",
		options = {
			{ "Debug: Skip", complete },
		},
	}
	local task = missionutil.makeTask(info)
		
	task.init = function(taskManager, data)
		taskManager:setMarker("cristal1", {entity = taskParams.cristal1pos, type = "question"})
		taskManager:setZone("trackEndAltdorf", { polygon = missionutil.makeCircleZone(taskParams.trackEndAltdorf, 150), draw = true, drawColor = missionutil.colors.GREEN})		
		taskManager:setZone("stationTunnel", { polygon = missionutil.makeCircleZone(game.interface.getEntity(taskParams.stationTunnel).position, 150), draw = true, drawColor = missionutil.colors.GREEN})	
		taskManager:guiSetEnabled("menu.construction.road.passenger", true)
	end
	
	task.update = function(taskManager)
		if task.isCompleted() then return end
				
		if game.interface.findPath(taskParams.stationConstructionSite, taskParams.stationAltdorf, { TRAIN = true }) then
			complete(taskManager, task)
		end
		
	end
	
	return task
end


local function makeStartConstructionTask(taskParams, doneFn)

	local function complete(taskManager, task)
		task.setCompleted()
		missionutil.startTask(taskManager,"buildTracks")
		game.interface.upgradeConstruction (taskParams.constructionSiteConstr, taskParams.tunnelIndustryConstrName, { productionLevel = 0, speed = 0, progress = 9, conversion = 0 })
		local strikeTask = taskManager:getById("strike")
		if strikeTask ~= nil then
			strikeTask.getUserState().pay = nil
		end
		
		local tunnelConstrData = game.interface.getEntity(taskParams.constructionSiteConstr)
		local tunnelData = game.interface.getEntity(tunnelConstrData.simBuildings[1])
		
	end
	
	local function startNewTechnique(taskManager, task)
		missionutil.startTask(taskManager, "newTechnique")
		return true
	end
	
	local function startStrike(taskManager, task)
		missionutil.startTask(taskManager, "strike")
		return true
	end
	
	local info = {
		title = _("Der Schutt muss weg!"),
		name = _("Stein abtransportieren"),
		paragraphs = {
			{ text = _("Mensch und Maschinen leisten gute Arbeit, die ersten Meter sind bereits geschafft. Im Berg drin ist es heiss wie in der Hölle und die Luft so dick, dass man sie mit dem Messer zerschneiden könnte.\n\n\z 
			Aus diesem Schlund muss der ganze Schutt nun raus. Und das schnell! Die Geldgeber beobachten unsere Fortschritte mit Argwohn...")
			}, 
			{ type = "TASK", text = _("Liefere Stein von der [link=${place}]Tunnelbaustelle[/link] zur [link=${place2}]Steinhalde[/link].") % { place = taskParams.constructionSiteConstr, place2=taskParams.dumpRockConstr }
			},
			{ type = "HINT", text = _("Kaufe einen Güterzug in [link=${depot}][/link], der Stein transportieren kann.") % { depot = taskParams.depot }
			},	
			{ type = "HINT", text = _("Erstelle eine Linie zwischen den zwei Güterbahnhöfen und setze den Zug darauf ein.")
			},	
			{ type = "HINT", text = _("Für jedes Werkzeug, das Du anlieferst, produziert die Tunnelbaustelle Stein, der wiederum abtransportiert werden muss.")
			},	
        },
		camera = { game.interface.getEntity(taskParams.dumpRockConstr).position[1], game.interface.getEntity(taskParams.dumpRockConstr).position[2], 1000.0 },
		voiceOver = "europe_01_05.wav",
		options = {
			{ "Debug: Skip", complete },
			{ "Debug: Start New Technique", startNewTechnique },
			{ "Debug: Start Strike", startStrike },
		},
	}
	local task = missionutil.makeTask(info)
		
	task.init = function(taskManager, data)
		task.getUserState().malus = 0
	end
	
	task.update = function(taskManager)
		if task.isCompleted() then return end
		
		local dumpConstrData = game.interface.getEntity(taskParams.dumpRockConstr)
		local tunnelConstrData = game.interface.getEntity(taskParams.constructionSiteConstr)
		
		local dumpData = game.interface.getEntity(dumpConstrData.simBuildings[1])
		local tunnelData = game.interface.getEntity(tunnelConstrData.simBuildings[1])
		
		task.setProgressCount(dumpData.itemsConsumed._sum, taskParams.stoneGoal + task.getUserState().malus)
		
		if (dumpData.itemsConsumed._sum >= taskParams.stoneGoal + task.getUserState().malus ) then
			complete(taskManager, task)
		elseif (dumpData.itemsConsumed._sum > taskParams.stoneGoal*0.5) then
			startStrike(taskManager, task)
		elseif (dumpData.itemsConsumed._sum > taskParams.stoneGoal*0.02) then
			startNewTechnique(taskManager, task)
		end
		
		local dumpUpdateStep = math.min(taskParams.stoneDumpUpdateSteps-1,math.floor(dumpData.itemsConsumed._sum/taskParams.stoneGoal*taskParams.stoneDumpUpdateSteps))
		if (dumpUpdateStep ~= dumpConstrData.params.heapsize ) then
			game.interface.upgradeConstruction (taskParams.dumpRockConstr, taskParams.dumpRockConstrName, { productionLevel = 0, heapsize = dumpUpdateStep })
		end
		
		local tunnelUpdateStep = math.min(taskParams.tunnelUpdateSteps-1,math.floor(dumpData.itemsConsumed._sum/taskParams.stoneGoal*taskParams.tunnelUpdateSteps))
		if (tunnelUpdateStep ~= tunnelConstrData.params.progress ) then
			game.interface.upgradeConstruction (taskParams.constructionSiteConstr, taskParams.tunnelIndustryConstrName, { productionLevel = tunnelConstrData.params.productionLevel, speed = tunnelConstrData.params.speed, conversion = tunnelConstrData.params.conversion, progress = tunnelUpdateStep })
		end

	end

	return task
end

local function makeNewTechniqueTask(taskParams, doneFn)
	
	local function complete(taskManager, task)
		task.setCompleted()
		missionutil.startTask(taskManager,"newTechniqueDone")
		game.interface.upgradeConstruction (taskParams.machinesFactoryConstr, taskParams.machinesFactoryConstrName, { productionLevel = 0, prodState = 0 })
		
		local tunnelConstrData = game.interface.getEntity(taskParams.constructionSiteConstr)
		local tunnelData = game.interface.getEntity(tunnelConstrData.simBuildings[1])
		
		game.interface.upgradeConstruction (taskParams.constructionSiteConstr, taskParams.tunnelIndustryConstrName, { productionLevel = tunnelConstrData.params.productionLevel, speed = tunnelConstrData.params.speed, conversion = 1, progress = tunnelConstrData.params.progress })

		task.addMedal("MEDAL_MACHINES")
	end
	
	local info = {
		title = _("Neue Technik!"),
		name = _("Werkzeuge liefern"),
		paragraphs = {
			{ text = _("Tag und Nacht wird gebohrt, gesprengt, geschuttet – doch trotzdem sind wir arg im Hintertreffen. Wir hinken dem Zeitplan eklatant hinterher und haben das ursprüngliche Budget bereits um das Zweifache überzogen. Der Bundesrat hat unseretwegen sogar schon eine Krisensitzung einberufen...\n\n\z 
			Wir brauchen dringend leistungsfähigere Bohrmaschinen, die nicht ständig kaputtgehen. Wir sollten eine entsprechende Werkstätte errichten. Das würde unser Vorhaben enorm beschleunigen!") 
			}, 
			{ type = "TASK", text = _("Liefere Werkzeuge zur [link=${place}]Werkstätte[/link].")  % { place = taskParams.machinesFactoryConstr }
			},
			{ type = "HINT", text = _("Die benötigte Menge an Werkzeugen ist sehr gering. Setze darum Pferdefuhrwerke ein!")  % { place = taskParams.machinesFactoryConstr }
			},
        },
		camera = { game.interface.getEntity(taskParams.machinesFactoryConstr).position[1], game.interface.getEntity(taskParams.machinesFactoryConstr).position[2], 1000.0 },
		voiceOver = "europe_01_06.wav",
		medalsHint = 1,
		options = {
			{ "Debug: Skip", complete },
		},
	}
	local task = missionutil.makeTask(info)
		
	task.init = function(taskManager, data)
		game.interface.upgradeConstruction (taskParams.machinesFactoryConstr, taskParams.machinesFactoryConstrName, { productionLevel = 0, prodState = 1 })
	end
	
	task.update = function(taskManager)
		if task.isCompleted() then return end
		
		local factoryConstrData = game.interface.getEntity(taskParams.machinesFactoryConstr)
		local factoryData = game.interface.getEntity(factoryConstrData.simBuildings[1])
		task.setProgressCount(factoryData.itemsConsumed._sum, taskParams.machinesFactoryGoal)
		
		if (factoryData.itemsConsumed._sum >= taskParams.machinesFactoryGoal) then
			complete(taskManager, task)
		end
		
	end

	return task
end


local function makeNewTechniqueDoneTask(taskParams, doneFn)
	
	local function complete(taskManager, task)
		task.setCompleted()
	end
	
	local info = {
		title = _("Neue Bohrmaschinen!"),
		name = _("Neue Bohrmaschinen verfügbar"),
		paragraphs = {
			{ text = _("Grossartig! Die neuen Bohrmaschinen sind ein Segen. Gingen früher ganze drei Maschinen in nur einem Tag kaputt, ist es heute nur noch eine in drei Tagen! Der Bundesrat ist hocherfreut. Oder nun ja, vielleicht eher: Etwas weniger ungehalten...") % { place = taskParams.machinesFactoryConstr }
			}, 
			{ type = "HINT", text = _("Dank der neuen Bohrmaschinen produziert die Tunnelbaustelle mehr Stein pro geliefertem Werkzeug.") 
			},
        },
		camera = { game.interface.getEntity(taskParams.machinesFactoryConstr).position[1], game.interface.getEntity(taskParams.machinesFactoryConstr).position[2], 1000.0 },
		voiceOver = "europe_01_07.wav",
		options = {
			{ _("Hervorragend!"), complete },
		},
	}
	local task = missionutil.makeTask(info)
		
	task.setProgressNone()
	return task
end

local function makeStrikeTask(taskParams, doneFn)
	
	local function addMonth(d)
		d.month = d.month+1
		if d.month > 12 then
			d.month = 1
			d.year = d.year+1
		end
	end
	
	local function isBefore(a, b)
		if a.year < b.year then
			return true
		elseif a.year == b.year then
			if a.month < b.month then
				return true
			elseif a.month == b.month then
				if a.day < b.day then
					return true
				end
			end
		end
		return false
	end
	
	local function useForce(taskManager, task)
		task.setCompleted()
		missionutil.startTask(taskManager,"messageUseForce")
		
		local tunnelConstrData = game.interface.getEntity(taskParams.constructionSiteConstr)
		local tunnelData = game.interface.getEntity(tunnelConstrData.simBuildings[1])
		
		game.interface.upgradeConstruction (taskParams.constructionSiteConstr, taskParams.tunnelIndustryConstrName, { productionLevel = tunnelConstrData.params.productionLevel, speed = 1, conversion = tunnelConstrData.params.conversion, progress = tunnelConstrData.params.progress })
		
		game.interface.setGameSpeed(1)
	end
	
	local function makeVent(taskManager, task)
		task.setCompleted()
		taskManager:getById("startConstruction").getUserState().malus = 100
		task.addMedal("MEDAL_PEACEFUL")
		task.addMedal("MEDAL_PEACEFUL")
		local tunnelConstrData = game.interface.getEntity(taskParams.constructionSiteConstr)
		local tunnelData = game.interface.getEntity(tunnelConstrData.simBuildings[1])
		
		game.interface.upgradeConstruction (taskParams.constructionSiteConstr, taskParams.tunnelIndustryConstrName, { productionLevel = tunnelConstrData.params.productionLevel, speed = 2, conversion = tunnelConstrData.params.conversion, progress = tunnelConstrData.params.progress })
		
		game.interface.setGameSpeed(1)
	end
		
	local function pay(taskManager, task)
		task.setCompleted()
		task.getUserState().pay = true
		task.addMedal("MEDAL_PEACEFUL")
		
		local tunnelConstrData = game.interface.getEntity(taskParams.constructionSiteConstr)
		local tunnelData = game.interface.getEntity(tunnelConstrData.simBuildings[1])
		
		game.interface.upgradeConstruction (taskParams.constructionSiteConstr, taskParams.tunnelIndustryConstrName, { productionLevel = tunnelConstrData.params.productionLevel, speed = 2, conversion = tunnelConstrData.params.conversion, progress = tunnelConstrData.params.progress })
		
		game.interface.setGameSpeed(1)
	end
	
	local info = {
		name = _("Streik!"),
		paragraphs = {
			{ text = _("Die Arbeiter hausen in Schweineställen und von den 3 Franken 90 Rappen Tageslohn müssen sie nicht nur Verpflegung, Unterkunft und Ausrüstung, sondern auch noch das Öl für ihre Lampen bezahlen. Inbegriffen, um sich die eigene Lampe zu füllen, ist immerhin täglich ein Liter Wein. Das ist ihnen offenbar nicht genug. Sie blockieren den Tunneleingang und werden ihn erst freigeben, wenn sie täglich einen Franken mehr erhalten. Was tun?\n\n\z
			Wenn wir uns das nicht leisten wollen, können wir die Polizei aus Uri anheuern und die Aufrührer niederknüppeln – oder wir kommen ihnen wenigstens ein Stück weit entgegen und sorgen für eine bessere Tunnelbelüftung.") 
			}, 
			{ type = "HINT", text = _("Wenn Du Gewalt anwendest, reduziert sich die Geschwindigkeit der Tunnelbaustelle um ein Drittel.") 
			},
			{ type = "HINT", text = _("Eine bessere Tunnelbelüftung hat zur Folge, dass ${malus} Einheiten Stein zusätzlich abtransportiert werden müssen.") % { malus = 100 } 
			},
			{ type = "HINT", text = _("Wenn Du die Arbeiter bezahlst, entstehen fortan monatliche Kosten von ${amount}.") % { amount=missionutil.makeReadableNumber(taskParams.strikePayAmount) } },
        },
		voiceOver = "europe_01_08.wav",
		medalsHint = 1,
		options = {
			{ _("Gewalt anwenden"), useForce },
			{ _("Belüftung verbessern"), makeVent }, 
			{ _("Bezahlen"), pay }
		},
	}
	local task = missionutil.makeTask(info)
		
	task.init = function(taskManager, data)
		
		local tunnelConstrData = game.interface.getEntity(taskParams.constructionSiteConstr)
		local tunnelData = game.interface.getEntity(tunnelConstrData.simBuildings[1])
		
		game.interface.upgradeConstruction (taskParams.constructionSiteConstr, taskParams.tunnelIndustryConstrName, { productionLevel = tunnelConstrData.params.productionLevel, speed = 0, conversion = tunnelConstrData.params.conversion, progress = tunnelConstrData.params.progress })
		
		task.getUserState().payday = game.interface.getGameTime().date
		addMonth(task.getUserState().payday)
	end
	
	
	task.update = function(taskManager)
		if task.getUserState().pay then
			local now = game.interface.getGameTime().date
			if isBefore(task.getUserState().payday, now) then
				task.getUserState().payday = now
				addMonth(task.getUserState().payday)
				game.interface.book(-taskParams.strikePayAmount)
			end
		end		
		if task.isCompleted() then return end
		game.interface.setGameSpeed(0)
	end

	return task
end


local function makeBuildTracks(taskParams, doneFn)

	local function complete(taskManager, task)
		taskManager:setZone("stationLuzern")
		taskManager:setZone("trackEndLuzern")
		taskManager:setZone("portalTunnel")
		taskManager:setZone("stationBellinzona")
		task.setCompleted()
		taskManager:add("useTunnel")
	end

	local info = {
		title = _("Strecke nach Bellizona"),
		name = _("${name1} und ${name2} verbinden") % { name1 = game.interface.getEntity(taskParams.towns[1]).name, name2 = game.interface.getEntity(taskParams.towns[3]).name },
		paragraphs = {
			{ text = _("Es ist soweit. Der Durchbruch ist geschafft! Ein Wunder, dass wir trotz der ganzen Strapazen nicht schon vor dem Durchstich tot im Tunnel umgekippt sind...\n\n\z
			Nun müssen wir nur noch die Eisenbahnstrecke fertigstellen, dann ist unsre Gotthardbahn perfekt!") 
			}, 
			{ type = "TASK", text = _("Vollende die Bahnstrecke von [link=${town0}][/link] nach [link=${town1}][/link].") %
				{ 
					 town0 = taskParams.towns[1], 
					 town1 = taskParams.towns[3]
				},
			},	
			{ type = "HINT", text = _("Schliesse die Lücken zwischen den markierten Gleisenden entlang der Strecke.")
			},
        },
		camera = { (game.interface.getEntity(taskParams.towns[1]).position[1] + game.interface.getEntity(taskParams.towns[3]).position[1]) / 2, (game.interface.getEntity(taskParams.towns[1]).position[2] + game.interface.getEntity(taskParams.towns[3]).position[2]) / 2, 3000.0 },
		voiceOver = "europe_01_09.wav",
		options = {
			{ "Debug: Skip", complete },
		},
	}	
		
	local task = missionutil.makeTask(info)
	
	
	task.init = function(taskManager, data)
		taskManager:setZone("Gotthard")
		taskManager:setZone("stationLuzern", { polygon = missionutil.makeCircleZone(game.interface.getEntity(taskParams.stationLuzern).position, 150), draw = true, drawColor = missionutil.colors.GREEN})		
		taskManager:setZone("trackEndLuzern", { polygon = missionutil.makeCircleZone(taskParams.trackEndLuzern, 150), draw = true, drawColor = missionutil.colors.GREEN})		
		
		taskManager:setZone("portalTunnel", { polygon = missionutil.makeCircleZone(taskParams.zonePortal, 150), draw = true, drawColor = missionutil.colors.RED})		
		taskManager:setZone("stationBellinzona", { polygon = missionutil.makeCircleZone(game.interface.getEntity(taskParams.stationBellinzona).position, 150), draw = true, drawColor = missionutil.colors.RED})		
	end
			
	task.update = function(taskManager)
		if task.isCompleted() then return end
		
		if missionutil.checkTownsConnected(taskParams.towns, "RAIL", { TRAIN = true }) then
			complete(taskManager, task)
		end
	
	end
	
	return task
end


local function makeUseTunnelTask(taskParams)
	
	local info = {
		title = _("Der erste Zug"),
		name = _("Zug durch den Tunnel schicken"),
		paragraphs = {
			{ text = _("Die Eisenbahnstrecke ist fertig. Nun ist es Zeit, uns mit dem Ruhm zu bekleckern, der uns für dieses Wunderwerk gebührt!\n\n\z
			Der feierlichen Einweihung steht nichts mehr im Wege. Also hopp und ab die Post!") %
				{ 
					 town0 = taskParams.towns[1], 
					 town1 = taskParams.towns[2],
					 town2 = taskParams.towns[3]
				},
			}, 
			{ type = "TASK", text = _("Schicke den ersten Zug von ${town0} über ${town1} nach ${town2}.") %
				{ 
					 town0 = game.interface.getEntity(taskParams.towns[1]).name, 
					 town1 = game.interface.getEntity(taskParams.towns[2]).name,
					 town2 = game.interface.getEntity(taskParams.towns[3]).name
				},
			},	
			{ type = "HINT", text = _("Kaufe im Depot einen Zug mit Passagierwaggons und lasse ihn auf einer Linie fahren, die in ${town0}, ${town1} und ${town2} Halt macht.") %
				{ 
					 town0 = game.interface.getEntity(taskParams.towns[1]).name, 
					 town1 = game.interface.getEntity(taskParams.towns[2]).name,
					 town2 = game.interface.getEntity(taskParams.towns[3]).name
				}
			},
			{ type = "HINT", text = _("Es bietet sich an, vor der Fahrt des Einweihungszugs die Gleise frei zu machen. Platziere passende Signale oder schicke andere Züge vorher ins Depot.") %
				{ 
					station = taskParams.finalStation
				}
			},
        },
		voiceOver = "europe_01_10.wav",
		options = {
			{ "Debug: Überspringen", function (taskManager) taskManager:getById("useTunnel").setCompleted() taskManager:add("messageMissionComplete") end }, 
		},
	}

	local task = missionutil.makeTask(info)
	
	task.getUserState().trains = { }
	task.getUserState().progress = { count = 0 }
	
	task.update = function (taskManager)
		if task.isCompleted() then return end
	
		local trains = task.getUserState().trains
		local progress = task.getUserState().progress
		
		local trainList = game.interface.getVehicles({ carrier = "RAIL" })
		
		for i = 1, #trainList do
			local vehicle = game.interface.getEntity(trainList[i])
			
			if (vehicle.state == "AT_TERMINAL") then
				local line = game.interface.getEntity(vehicle.line)
				local stationGroup = game.interface.getEntity(line.stops[vehicle.stopIndex + 1])
				local station = game.interface.getEntity(stationGroup.stations[1])
				
				local n = (trains[trainList[i]] or 0) + 1
				
				assert(n < #taskParams.towns + 1)

				if stationGroup.stations[1] == taskParams.towns[n] or station.town == taskParams.towns[n] then
					trains[trainList[i]] = n
					
					task.getUserState().progress.count = math.max(progress.count, n)
				end
			end
		end
		
		task.setProgressCount(math.min(progress.count, #taskParams.towns), #taskParams.towns)
		
		if  progress.count >= #taskParams.towns then
								
			task.setCompleted()
			taskManager:add("messageMissionComplete")
		end
		
	end

	return task
end


local function makeCristal1(taskParams)

	local info = {
		name = _("Der sagenumwobene Bergkristall (1/3)"),
		paragraphs = {
			{ text = _("Hoch oben in den Bergen, so erzählt man es sich hier seit langer Zeit, soll es einen Bergkristall von unvorstellbarer Schönheit geben. Ein alter Alphirte kennt angeblich die ganze Geschichte...") % { town = taskParams.cristal2pos },
			}, 
			{ type = "TASK", text = _("Folge dem Hinweis nach [link=${town}][/link] und klicke auf das Fragezeichen, um mehr zu erfahren.") %
				{ 
					town = taskParams.cristal2pos 
				},
			},	
			{ type = "HINT", text = _("Mithilfe des Fadenkreuzes am linken, oberen Rand des Aufgabenfensters kannst Du zur Gegend springen, wo der Kristall angeblich verborgen ist. Suche nach weiteren Hinweisen, um genauere Angaben zu bekommen.") }
        },
		camera = { 1500.0, 1800.0, 2500.0 },
		voiceOver = "europe_01_11.wav",
		medalsHint = 1,
	}
		
	local task = missionutil.makeTask(info)
		
	task.init = function(taskManager, data)
		taskManager:setMarker("cristal2", {entity = taskParams.cristal2pos, type = "question"})
	end
		
	task.update = function (taskManager)
		if task.isCompleted() then return end
	end
	
	return task
end


local function makeCristal2(taskParams)

	local info = {
		name = _("Der sagenumwobene Bergkristall (2/3)"),
		paragraphs = {
			{ text = _("Der alte Mann sagt, er kenne den Kristall nur aus Sagen und Legenden, doch seine junge Enkelin, die treibe sich mit ihren Geissen oft hoch oben im Gebirge rum...\n\n\z
			Das Mädchen kraxelt durch die Felsen wie eine kleine Bergziege. Man kommt ihr kaum hinterher...") ,
			}, 
			{ type = "TASK", text = _("Folge dem Hinweis und klicke auf das Fragezeichen, um mehr zu erfahren.") },
			{ type = "HINT", text = _("Mithilfe des Fadenkreuzes am linken, oberen Rand des Aufgabenfensters kannst Du zur Gegend springen, wo der Kristall angeblich verborgen ist. Suche nach weiteren Hinweisen, um genauere Angaben zu bekommen.") }
        },
		camera = { 1500.0, 1800.0, 1500.0 },
		voiceOver = "europe_01_12.wav",
	}
		
	local task = missionutil.makeTask(info)
		
	task.init = function(taskManager, data)
		taskManager:setMarker("cristal3", {pos = taskParams.cristal3pos, type = "question"})
		taskManager:getById("taskCristal1").setCompleted()
	end
		
	task.update = function (taskManager)
		if task.isCompleted() then return end
	end
	
	return task
end


local function makeCristal3(taskParams)

	local info = {
		name = _("Der sagenumwobene Bergkristall (3/3)"),
		paragraphs = {
			{ text = _("Hier also soll es irgendwo sein. Eine besondere Felsformation weise auf den sagenhaften Bergkristall hin, hat uns das Mädchen noch zugerufen, bevor es flink davon geklettert ist."),
			}, 
			{ type = "TASK", text = _("Drehe jeden Stein der Felsformation um, dann findest Du den Bergkristall.") },	
			{ type = "HINT", text = _("Mithilfe des Fadenkreuzes am linken, oberen Rand des Aufgabenfensters kannst Du zur Gegend springen, wo der Kristall angeblich verborgen ist. Suche nach weiteren Hinweisen, um genauere Angaben zu bekommen.") }
        },
		camera = { 1380.0, 1673.0, 700.0 },
		voiceOver = "europe_01_13.wav",
	}
		
	local task = missionutil.makeTask(info)
	
	task.init = function(taskManager, data)
		taskManager:getById("taskCristal2").setCompleted()
	end
	
	
	task.update = function (taskManager)
		if task.isCompleted() then return end
				
		local rocks = game.interface.getEntities({pos = {1380, 1673}, radius = 10})
		
		if #rocks < 2 then
			task.setCompleted()
			taskManager:add("taskCristal4")
		end
	end
	
	return task
end


local function makeCristal4(label)
	
    local info = {
        name = _("Bergkristall gefunden!"),
        description = _("Was für ein wundervolles und seltenes Exemplar! Der Kristall strahlt eine ungewöhnliche Energie aus."),
		voiceOver = "europe_01_14.wav",
    }
	
	local function fn(taskManager, task)		
		task.setCompleted()
		task.addMedal("MEDAL_CRISTAL")
	end

	info.options = { { _("Juhu!"), fn } }
	info.optionsRightAlign = true

	local task = missionutil.makeTask(info)

	task.setProgressNone()

	return task
end


local function makeMissionComplete(taskParams, label)

	local info = {
		title = _("Mission gewonnen!"), 
		name = _("Geschafft!"),
		description = _("Freude herrscht! Die neue Achse zwischen Nord und Süd ist eröffnet. Eine Meisterleistung der Ingenieurskunst! Bravo!"),
		voiceOver = "europe_01_15.wav",
	}

	local function fn(taskManager, task)
	
		if (taskManager:getById("taskMoney").getUserState().current == 1) then
			task.addMedal("MEDAL_RICH")
		end
				
		missionutil.completeAllTasks(taskManager)
	end

	local task = missionutil.makeMissionCompleteTask(info, fn)

	task.init = function(taskManager, data)
		if (game.interface.getGameTime().date.year <= taskParams.fastYear) then
			task.addMedal("MEDAL_FAST")
		end
	end

	return task
end

local function getRadFromVec(vec2)
	vec2.y = vec2.y * -1
	local sqrt = math.sqrt
	local coordVec = {x = 0, y = -1}
	local xW, yW, zW = coordVec.x, coordVec.y, coordVec.z;
	local xH, yH, zH = vec2.x, vec2.y, vec2.z;
	local rad = math.acos(
						math.abs(vec2.x * coordVec.x + vec2.y * coordVec.y)
	            		/ 
	            		( 
	            			sqrt(vec2.x * vec2.x + vec2.y * vec2.y) 
	            			* 
	            			sqrt(coordVec.x * coordVec.x + coordVec.y * coordVec.y)
	            		)
	            	);

	if (vec2.x < 0 and vec2.y <= 0) then
		return rad 
	elseif (vec2.x <= 0 and vec2.y > 0) then
		return rad + math.rad(90)
	elseif (vec2.x > 0 and vec2.y >= 0) then
		return rad + math.rad(180)
	elseif (vec2.x >= 0 and vec2.y < 0) then
		return rad + math.rad(270)
	end
	print("error")
	return rad
	

end

local function printObj(obj, hierarchyLevel) 
  if (hierarchyLevel == nil) then
      hierarchyLevel = 0
  end

  local whitespace = ""

  for i=0,hierarchyLevel,1 do
      whitespace = whitespace .. "\t"
  end
  whitespace = whitespace .. ":"
  	if(obj ~= nil) then
		for k,v in pairs(obj) do 
	  		print(whitespace, k, v)
	  	
	  		if (v == "params") then

	  		elseif (type(v) == "table") then
		  		printObj(v, hierarchyLevel+1)
			end
			if(k == "simBuildings" and v[1] ~= nil) then
				printObj(game.interface.getEntity(v[1]), hierarchyLevel + 1)
			end  
	  	end
	end
  
       
end

local function makeStationTask( taskParams, doneFn ) 
	local function complete(taskManager, task)
		task.setCompleted()
	end

	--Erzeugt ein Infor Objekt mit den Werten: title, name, paragraphs
	local info = {
		title = _("Baubeginn"),
		name = _("Stahl liefern"),
		paragraphs = {
			{ text = _("Baue einen Bahnhof!")
			}, 
			{ type = "TASK", text = _("Baue einen Bahnhof") --[link=${steelMill}][/link] zur [link=${place}]Baustelle[/link]") % { steelMill = taskParams.steelMill, place = taskParams.constructionSiteConstr }
			},		
			{ type = "HINT", text = _("Stelle den Bahnhof einfach ab.") 
			},
		}
	}
	--erstellt den Task mit den Obigen Infos
	local task = missionutil.makeTask(info)

	task.init = function(taskManager, data)
		
	end
	
	task.update = function(taskManager)
		jj = jj + 1;
		if jj >= 1000 then
			--printObj(taskParams)
			
			print(" ------ Export", expcount, " ------ ")
			expcount = expcount + 1
			--print(taskParams.oelraf)
			local nearBuildings = game.interface.getEntities({pos = {0, 0}, radius = 900000}, { type = "CONSTRUCTION"})
			for i=1,#(nearBuildings) do
				local buildingId = nearBuildings[i];
				local building = game.interface.getEntity(buildingId)

				if building.simBuildings[1] ~= nil then
					--print("##############start###################")
					--print("x:", building.transf[2], "y:", building.transf[1])
					--print("rad:", )
					local rad = getRadFromVec({x = building.transf[2], y = building.transf[1]})
					local fileName = building.fileName
					local pos = {x = building.position[1], y = building.position[2]}

					print("{ pos = {", pos.x, ",", pos.y, "}, angle = ", rad, ", fileName =\"" .. fileName .. "\"},")

					--print("##############endee###################")
				end
			end
			print(" ------ Export", "ende", " ------ ")
			
			

			jj = 0;
		end
		
		--print(game.interface.stations)
		--local allStations =game.interface.getStations();
		--if allStations ~= nil then
		--	printObj(allStations)
		--end

	end
	return task
end

function data()
	local tm = TaskManager.new()

	
	local lockDownGotthard = {  }
	for i = 200, -200, -1 do
		lockDownGotthard[#lockDownGotthard + 1] = {i*10, 0}
	end
	for i = -200, 200 do
		lockDownGotthard[#lockDownGotthard + 1] = {i*10, -875 }
	end
	
	local tunnelIndustryConstrName 	= "industry/tunnelIndustry_new.con"
	
	local dumpRockConstrName 			= "industry/stone_dump.con"
	local foodProcessingPlantConstrName = "industry/food_processing_plant.con"
	local machinesFactoryConstrName 	= "industry/machines_factory.con"
	local toolsFactoryConstrName 		= "industry/tools_factory.con"
	local constructionMaterialConstrName = "industry/construction_material.con"
	
	local gotthardData = {
		tunnelUpdateSteps			= 9,
		passengersGoal				= 100,
		gotthardPos					= {100, -89},
		
		machinesFactoryConstrName	= machinesFactoryConstrName,
		tunnelIndustryConstrName 	= tunnelIndustryConstrName,
		dumpRockConstrName			= dumpRockConstrName,
		toolsFactoryConstrName		= toolsFactoryConstrName,
		constructionMaterialConstrName		= toolsFactoryConstrName,
		constructionSite			= 7624,
				
		stationAltdorf 				= 9241,
		stationAltdorfConstr		= 9240,
		
		stationLuzern				= 9440,
		stationLuzernConstr			= 9438,
		
		stationTunnel				= 7613,
		stationTunnelConstr			= 7647,
		
		stationConstructionSite		= 7613, 
		stationConstructionSiteConstr = 7647,
		
		stationDump					= 9212,
		stationDumpConstr 			= 2642,
		
		toolsFactory				= 9716,
		toolsFactoryConstr			= 8940,
		
		stationBellinzona			= 8522,
		stationBellinzonaConstr		= 8526,
		
		cristal1pos 				= 7624,
		cristal2pos 				= 7682,
		cristal3pos 				= {930, 1590, 363},
		
		depot 						= 9379,
		
		zonePortal 					= {200, -870},
		trackEndAltdorf 			= {455, 855},
		trackEndLuzern				= {-472, 2137},
				
		rocks = { 1568, 2714, 9539, 9685, 9699},
				
		towns = {877, 7682, 8158}
	}
	--print("GAMEGAMEGAMEGAMEGAMEGAMEGAMEGAMEGAMEGAMEGAMEGAMEGAMEGAMEGAMEGAME")
	--local stations = game.interface.getEntities({ type="STATION_GROUP" })
	--for i = 1, #(stations) do
	--	printObj(game.interface.getEntity(stations[i]))
	--end
	--print("ENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDE")
	
	--local tns = game.interface.getTowns()
	--for i=1, #tns do
	--	print("TOWNSTOWNSTOWNSTOWNSTOWNSTOWNSTOWNSTOWNSTOWNSTOWNSTOWNSTOWNSTOWNS")
	--	printObj(game.interface.getEntity(tns[i]))
	--	print("ENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDE")
	--end

	
	
	local allConstructions = game.interface.getEntities({ pos = {0,0}, radius = 900000 }, { type="CONSTRUCTION", name })
	

	for i=1, #allConstructions do
		local c = allConstructions[i]
		local cData = game.interface.getEntity(c)
		if cData.simBuildings[1] ~= nil then
			--print("INDUSTRYINDUSTRYINDUSTRYINDUSTRYINDUSTRYINDUSTRYINDUSTRYINDUSTRY")
			--print(cData.id)
			--print("ENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDEENDE")
			if cData.name == "Selberting Ölraffinerie" then
				gotthardData.oelraf = cData.id
			end
		end
		if (cData.fileName == tunnelIndustryConstrName) then
			gotthardData.constructionSiteConstr = c
		elseif (cData.fileName == dumpRockConstrName) then
			gotthardData.dumpRockConstr = c
		elseif (cData.fileName == foodProcessingPlantConstrName) then
			gotthardData.foodProcessingPlantConstr = c
		elseif (cData.fileName == machinesFactoryConstrName) then
			gotthardData.machinesFactoryConstr = c
		elseif (cData.fileName == constructionMaterialConstrName) then
			--print("HERE-------------------------TEST")
			--printObj(cData)
			--printObj(game)
			--print("SONDER")
			--for key,value in pairs(cData.transf) do print(key,value) end
			--print("ENDE")
			--print("HERE-------------------------ENDE")
			gotthardData.constructionMaterialConstr = c
		end
	end

	
	
	local medalsData = {
		fastYear					= 1885,
	}
	
	tm:register("messageIntro", 					makeIntroTask, 							gotthardData,					{ missionutil.addTask("buildStation", tm) },	 	_("Start mission..."))
	tm:register("buildStation", 					makeStationTask, 						gotthardData)
	tm:register("messageMissionComplete", 			makeMissionComplete, 		 			medalsData, _("Done"))
	


	local taskMoney = { startBudget = 5000000, loanFactor = 1, loanTrigger = 250000 }	
	missionutil.registerMoneyTask(tm, taskMoney)
	
	local mission = missionutil.makeMissionInterface(tm)

	mission.onInit(function ()
		tm:add("messageIntro")
		tm:add("taskMoney")

		--tm:setZone("Gotthard", { polygon = lockDownGotthard, draw = true, drawColor = missionutil.colors.RED, buildToolMode = "PROHIBIT" })
		
	end)
	
	mission.onUpdate(function()

	end)
	
	mission.onMarkerEvent(function (params)
		
		--if tm:getById("messageMissionComplete") ~= nil then tm:setMarker(params.key) return end
	end)
	
	mission.onInitGui(function ()
		game.gui.playTrack("res/audio/music/06 - Left Side.ogg", 0.0)
	end)

	return mission
end
